//
//     WARNING
//       This code is completely open and available to be modified and changed
//       in any way and by anybody. It is yours once you start to change it.
//       Don't try to patent your derivative code and sue us for infringement.
//       We were here first!
//
//       The reason it is provided this way is for your education and to illustrate some coding
//       ideas. It can cause damage to your Arduino or other hardware if you are not extremely careful.
//       No guarantee is offered and no risk is assumed by the code author.
//       All the risk is on you even if you follow the commments provided in the
//       source code or elsewhere. This is powerful code but you must know it well.
//
//       This code implements the Command Module as imagined by Robot-DHQ. The main purpose of
//       the code is to organize and send commands to one or more Robot-DHQ modules.
//
//       Robot-DHQ is the Robot Design Headquarters
//
//       The human interface paradigm is "immediate response" to any input. No CR or LF is required if
//       the input is a key on a keyboard. Similarly a button press sends a command immediately and
//       the code does not wait for the button to be released. Electrically the button press could
//       be the low to high transition or the high to low transition. If the button is behaving
//       "wrongly" one option is to make a change to the code.
//
//       To view the command interpreter switch logic search for processCommand() or command_number.
//       To view the console display logic search for consoleRefresh()
//
//       It has many undocumented features and has not been optimized. You can learn
//       a lot by studying it or you can use it according to the documentation and
//       not read any further. Notes within the source code will give you some ideas
//       for using some of the undocumented features.
//
//       Consider it a "work in progress". The Command Module is described in
//       this document: D:\Projects\Arduino\Smart Modules\??? Command Module.odt
//
//       This code is designed to run on an Arduino Nano installed on a Robot-DHQ adapter board.
//       The serial port on the Arduino Nano accepts commands and displays the command menus.
//       The port is configured to operate an 115,200 baud.
//       This is the first thing to check if the menu is not readable when first connected.
//
//       The commands sent to Robot-DHQ modules are sent over a second serial interface.
//       The Arduino Nano only has one hardware serial interface so the second serial
//       interface is a software serial port. As a consequence there are a number of limitations.
//       One limitation is the need to operate this port at 57,600 baud, max.
//       Also, some delays are required in code to make the output readable.
//       Communication between modules is at 57,600 baud.
//
// how to avoid destroying the Arduino Nano
// 1   the Arduino will not be detected over the USB interface if the Schottky diode
//     has been destroyed. This can happen the +5 volt supply is accidentally
//     shorted to ground. The diode can only handle about 0.5 amps before it blows out.
//     This is therefore the first thing to suspect if the serial connection drops.
//     Therefore, always check for shorts before applying power
//
// 2   input pins are relatively safe to implement as long as they are not
//     connected to a power source with a voltage greater or less than the
//     hardware limitations, usually VCC and GROUND. The S1025 command is
//     provided to produce an input only state.
//

#include <SoftwareSerial.h>

SoftwareSerial gtSerial1(10, 13); // RX is Arduino pin D11, TX is Arduino pin D13
SoftwareSerial gtSerial2(12, 8); // RX is Arduino pin D12, TX is Arduino pin D8


// port allocations, etc.
const int NANO_05_D2 = 2;   // Socket C Pin 8, Socket E Pin 4
const int NANO_06_D3 = 3;   // Socket C Pin 5, Socket E Pin 5
const int NANO_07_D4 = 4;   // Socket C Pin 4, Socket E Pin 8
const int NANO_08_D5 = 5;   // Socket C Pin 1, Socket E Pin 7
const int NANO_09_D6 = 6;   // Socket C Pin 2, Socket B Pin 4, Socket D Pin 10
const int NANO_10_D7 = 7;   // Socket B Pin 5, Socket D Pin 9
const int NANO_11_D8 = 8;   // Socket B Pin 8, Socket D Pin 7
const int NANO_12_D9 = 9;   // Socket B Pin 7, Socket D Pin 1
const int NANO_13_D10 = 10; // Socket D Pin 5
const int NANO_14_D11 = 11; // Socket D Pin 2
const int NANO_15_D12 = 12; // Socket D Pin 8
const int NANO_16_D13 = 13; // Socket D Pin 4
const int NANO_19_A0 = A0;  // Socket A Pin 1, Socket E Pin 2
const int NANO_20_A1 = A1;  // Socket A Pin 4, Socket E Pin 1
const int NANO_21_A2 = A2;  // Socket A Pin 2, Socket E Pin 10
const int NANO_22_A3 = A3;  // Socket A Pin 10, Socket E Pin 9
const int NANO_23_A4 = A4;  // Socket A Pin 8, Socket B Pin 9
const int NANO_24_A5 = A5;  // Socket A Pin 9, Socket B Pin 1
const int NANO_25_A6 = A6;  // Socket A Pin 7, Socket B Pin 10
const int NANO_26_A7 = A7;  // Socket A Pin 5, Socket B Pin 2

// these variables affect the configuration settings at startup
//   timing parameters
//     the timing defaults can be adjusted by changing these parameters here
const int slow_loop_size = 50;
// implementing a pseudo millisecond timer
const unsigned long fast_loop_size = 1000;
// end of configuration constants

//   input and output routing
//     the output is serial (to robot and monitor) if the input is keyboard.
//     the output is serial (to robot and monitor) if the input is a button.
//     the output is serial (to monitor) if the input is serial (from robot)
//     the output is LED to indicate command status mode, etc.
//     internal digital settings may indicate an input set on a pot.

// timing variables
unsigned long startMicros = 0;
unsigned long begin_low = 0;
unsigned long begin_high = 0;
//unsigned long message_timer_start = 0;
int bit_t1 = 0;
int bit_t2 = 0;
int bit_t3 = 0;
int bit_t4 = 0;
int bit_t5 = 0;
int bit_t6 = 0;
int bit_t7 = 0;
int bit_num = 0;
int last_D11 = 1;
int new_message = 0;
int chars_in = 0;
int message_len;
int char_in;
int message_mode = 0;
const int BIT_PERIOD = 120;
const int END_CHAR_PERIOD = 500;
const int END_MSG_PERIOD = 30000;

// new variables

byte rx_byte_S2 = 0;        // stores received byte
char recd_char;
int commands_sent = 0;
int display_menu = 0;
int sequence_steps_to_do = 0;
int repeats_to_do = 0;
int repeating = 0;
// the default action is to immedeiately show all incoming data from the remote node
// this can be "skipped" when a sequence is collecting data for a report, etc.
int show_responses = 1; // show responses from remote node
int step_no = 0; // show responses from remote node
int view_sending = 0;
int menu_number = 0;
int cmd_menu = 0;
int seq_menu = 8;
int input_mode = 0;
int digit_key = 0;
// special parms
int pin_2_min = 0;
int pin_2_max = 0;
int pin_3_min = 0;
int pin_3_max = 0;
int pin_4_min = 0;
int pin_4_max = 0;
int pin_5_min = 0;
int pin_5_max = 0;
// settable parms
int digital_pin = 2;
int saved_digital_pin = 2;
int analog_pin = 0;
int parm_value_V = 0;
int parm_value_W = 0;
int parm_value_X = 0;
int parm_value_Y = 0;
int parm_value_Z = 0;
int temp_value = 0;
int entered_value = 0;
int digit_count = 0;
int h_idx = 0;
int p_idx = 0;
int data_pos_S1 = 0;
int data_pos_S2 = 0;
String myStr = "";
//String messageStr = "";
String last_command;
String tmpString;


const int NUMBER_OF_LINES = 5;
const int LINE_SIZE = 50;
char outputBuffer [NUMBER_OF_LINES] [LINE_SIZE];



const int NUMBER_OF_ELEMENTS = 16;
const int MAX_SIZE = 12;
char cmdHistory [NUMBER_OF_ELEMENTS] [MAX_SIZE];
char seqCommand [NUMBER_OF_ELEMENTS] [MAX_SIZE];
int seqDelay [NUMBER_OF_ELEMENTS];
char cmdResponse_S1[81];
char cmdResponse_S2[81];
char savResponse[20];
int sourcePin = 0;
int pulseWidth = 0;
int parmOne = 0;
int parmTwo = 0;

// the pin_ array variables configure and control the digital IO pins
//   pin_function indicates the current function:
//     0: raw two-state input
//     1: debounced input
//     2: LED output
//   pin_io is 0 if configured for input, 1 if configured for output
//   pin_state is the raw digital input
//   pin_last is used to debounce a pushbutton switch
//   pin_now is the digital input raw or debounced
//   pin_led_on_time is the on duration (if LED)
//   pin_counter measures pulse width (LED or PWM)

int pin_function[12];
int pin_io[12];
int pin_state[12];
int pin_last[12];
int pin_now[12];
int pin_led_on_time[12];
int pin_counter[12];

// indexes and counters
int pin_idx;
int led_idx = 0;
int slow_counter = 0;
int seq_pause_counter = 0;
int seq_step = 0;
int p_seq_step = 0;
int menu_option = 0;

void setup() {
  pinMode(11, INPUT);

  gtSerial1.begin(57600);  // connection to text source node
  gtSerial2.begin(57600);  // connection to servo node
  Serial.begin(57600);
  while (!Serial) {
    ; // wait for serial port to connect. Needed for native USB port only
  }

  initialize_digital_arrays();
  Serial.println(F("Command Module"));
  Serial.print(F("Bit Period = "));
  Serial.println(BIT_PERIOD);
  Serial.print(F("End Char Period = "));
  Serial.println(END_CHAR_PERIOD);
  Serial.print(F("End Msg Period = "));
  Serial.print(END_MSG_PERIOD);
  consoleRefresh();
  message_mode = 1; // clear the message buffer
  begin_high = micros();
}

void initialize_digital_arrays() {
  for (p_idx = 0; p_idx < 5; p_idx++) {
    myStr = "";
    myStr.toCharArray(outputBuffer[p_idx], 50);
  }
  for (h_idx = 0; h_idx < 14; h_idx++) {
    myStr = "";
    myStr.toCharArray(cmdHistory[h_idx], 12);
  }
  for (pin_idx = 0; pin_idx < 12; pin_idx++) {
    pin_state       [pin_idx] = 1;
    pin_last        [pin_idx] = 0;
    pin_now         [pin_idx] = 0;
    pin_counter     [pin_idx] = 0;
    pin_led_on_time[pin_idx] = 0;
  }
}

void loop() {
  if (message_mode > 0) {
    // begin message_mode
    // output of remote processors is buffered and displayed when a CR is received.
    if (digitalRead(11) < 1) {
      // D11 low
      if (last_D11 > 0) { // high to low transition
        begin_low = micros();
        if (bit_num < 1) {
          bit_num++; // start of char
        } else {
          if (bit_num < 3) {
            bit_t2 = begin_low - begin_high;
            if (bit_t2 > END_CHAR_PERIOD) {
              bit_num = 0;
              cmdResponse_S1[chars_in] = '*';
              if (chars_in < 80) {
                chars_in++;
              }
            } else {
              if (bit_t2 > BIT_PERIOD) {
                char_in += 32; // bit 2
              }
              bit_num++;
            }
          } else {
            if (bit_num < 5) {
              bit_t4 = begin_low - begin_high;
              if (bit_t4 > END_CHAR_PERIOD) {
                bit_num = 0;
                cmdResponse_S1[chars_in] = '*';
                if (chars_in < 80) {
                  chars_in++;
                }
              } else {
                if (bit_t4 > BIT_PERIOD) {
                  char_in += 8; // bit 4
                }
                bit_num++;
              }
            } else {
              if (bit_num < 7) {
                bit_t6 = begin_low - begin_high;
                if (bit_t6 > END_CHAR_PERIOD) {
                  bit_num = 0;
                  cmdResponse_S1[chars_in] = '*';
                  if (chars_in < 80) {
                    chars_in++;
                  }
                } else {
                  if (bit_t6 > BIT_PERIOD) {
                    char_in += 2; // bit 6
                  }
                  bit_num++;
                }
              }
            }
          }
        }
        last_D11 = 0;
      }
    } else {
      // D11 high
      if (last_D11 < 1) { // low to high transition
        begin_high = micros();
        if (bit_num < 2) {
          bit_t1 = begin_high - begin_low;
          if (bit_t1 > BIT_PERIOD) {
            char_in = 64; // bit 1
          } else {
            char_in = 0; // bit 1
          }
          bit_num++;
        } else {
          if (bit_num < 4) {
            bit_t3 = begin_high - begin_low;
            if (bit_t3 > BIT_PERIOD) {
              char_in += 16; // bit 3
            }
            bit_num++;
          } else {
            if (bit_num < 6) {
              bit_t5 = begin_high - begin_low;
              if (bit_t5 > BIT_PERIOD) {
                char_in += 4; // bit 5
              }
              bit_num++;
            } else {
              bit_t7 = begin_high - begin_low;
              if (bit_t7 > BIT_PERIOD) {
                char_in += 1; // bit 7
              }
              add_to_array();
              bit_num = 0;
            }
          }
        }
        last_D11 = 1;
      }
    }

    if (micros() - begin_high > END_MSG_PERIOD) {
      Serial.println();
      Serial.print(message_len);
      Serial.print(F(","));
      Serial.println(cmdResponse_S1);
      message_mode = 0;
      cmdResponse_S1[chars_in] = '\0';
      //      messageStr = "";
      //      messageStr.concat(cmdResponse_S1);
      //      Serial.print(F("s: "));
      //      Serial.println(messageStr);
    }
    // end message_mode
  } else {
    // begin not message_mode
    if (micros() - startMicros >= fast_loop_size) {
      // check clock and send queued and scheduled commands to remote processor.
      slowLoop();          // once per millisecond, 1000 times per second, period = 1000 microseconds
      startMicros = micros();
    }
    //    digitalWrite(led_idx + 2, HIGH);
    //    digitalWrite(led_idx + 2, LOW);

    // each character is processed as a complete command as it is entered (CR not required)
    while (Serial.available() > 0) {
      recd_char = Serial.read();
      processCommand();
      if (display_menu > 0) {
        // the "console" includes menus, prompts and help displays
        consoleRefresh();
        display_menu = 0;
      }
    }

    // output of remote processors is buffered and displayed when a CR is received.
    // check if byte available on the software serial port
    if (gtSerial2.available()) {
      // get the byte from the software serial port
      rx_byte_S2 = gtSerial2.read();
      processgtSerial2Data();
    }
    // end not message_mode
  }
}

void add_to_array() {
  if (new_message > 0) {
    new_message = 0;
    message_len = char_in;
    chars_in = 0;
  } else {
    if (13 == char(char_in)) {
      cmdResponse_S1[chars_in] = '\0';
      Serial.println();
      Serial.println(F("13 recd"));
    } else {
      cmdResponse_S1[chars_in] = char(char_in);
    }
    if (chars_in < 80) {
      chars_in++;
    }
  }
  message_mode = 1;
  begin_high = micros();
}

void slowLoop() { // millisecond
  slow_counter++;
  if (slow_counter > slow_loop_size) {
    slow_counter = 0;
    // these actions are done once per 50 millisecond, 20 times per second, period = 50,000 microseconds
    digital_reads();
  }
  if (sequence_steps_to_do > 0) {
    seq_pause_counter++;
    if (seq_pause_counter > seqDelay[seq_step]) { // pause before sending the next command
      seq_pause_counter = 0;
      p_seq_step = seq_step;
      executeCommand(seqCommand[seq_step]);
      if (p_seq_step > seq_step) { // jumping to a new sequence
        executeCommand(seqCommand[seq_step]);
      }
      seq_step++;
      sequence_steps_to_do--;
      if (sequence_steps_to_do < 1) {
        if (show_responses > 0) {
          Serial.println();
          Serial.println(F("Command sequence is complete."));
        } else {
          show_responses = 1; // show responses
        }
      }
    }
  }
}

// examine this code to understand how to manipulate sequences
// the first letter in the command string is key to what it does
void executeCommand(String command_string) {
  switch (command_string.charAt(0)) {
    case 'M': // execute a Menu option - this command can replace a sequence with a different one
      menu_option = 1;
      parmOne = command_string.charAt(1) - 48;
      parmOne = parmOne * 10 + command_string.charAt(2) - 48;
      parmTwo = command_string.charAt(3) - 48;
      parmTwo = parmTwo * 10 + command_string.charAt(4) - 48;
      if (parmOne > -1) {
        if (parmOne < 10) {
          if (parmTwo > -1) {
            if (parmTwo < 14) {
              // parm one is menu number (zero based)
              // parm two is option number (zero based)
              res_locator(parmOne, parmTwo, 1); // Warning: command may fail silently if arguments are invalid
            }
          }
        }
      }
      menu_option = 0;
      break;
    case '>': // remember the current digital pin
      saved_digital_pin = digital_pin;
      break;
    case '<': // restore the current digital pin
      digital_pin = saved_digital_pin;
      set_pin_digital();
      break;
    case 'P': // print report
      Serial.println();
      Serial.println();
      switch (command_string.charAt(1) - 48) {
        case 1:
          report_one();
          break;
        case 2:
          report_two();
          break;
      }
      break;
    case 'R': // return x steps
      parmOne = command_string.charAt(1) - 48;
      parmOne = parmOne * 10 + command_string.charAt(2) - 48;
      parmTwo = command_string.charAt(3) - 48;
      parmTwo = parmTwo * 10 + command_string.charAt(4) - 48;
      Serial.println();
      Serial.print(F("returning repeating = "));
      Serial.print(repeating);
      Serial.print(F(", parmTwo = "));
      Serial.print(parmTwo);
      Serial.print(F(", repeats_to_do = "));
      Serial.println(repeats_to_do);
      if (parmTwo > 0) {
        if (repeating > 0) {
          if (repeats_to_do < 1) {
            repeating = 0;
            terminateCurrentSeq();
          }
        } else {
          repeats_to_do = parmTwo;
          repeating = 1;
        }
      }
      repeats_to_do--;
      sequence_steps_to_do += parmOne;
      seq_step -= parmOne;
      break;
    case 'S':
      sendCommand(command_string);
      break;
    default:
      Serial.println(F("noop")); // not a valid sequence command
      break;
  }
}

// this function includes switch debounce logic
void digital_reads() {
  for (pin_idx = 0; pin_idx < 12; pin_idx++) {
    pin_state[pin_idx] = digitalRead(pin_idx + 2);
    if (pin_state[pin_idx] == pin_last[pin_idx]) {
      pin_now[pin_idx] = pin_state[pin_idx];
    } else {
      if (1 == pin_now[pin_idx]) {
        //  res_locator(menu_number, menu_line, 1); // res_type 1 to execute the command
        switch (pin_idx) {
          case 6:
            //            Serial.print(F("S1 pressed: "));
            //            Serial.print(pin_idx);
            //            if (2 == digital_pin) {
            //              res_locator(8, 4, 1); // execute sequence menu #8 item 4 alternate between servo extremes
            //            } else {
            //              res_locator(8, 6, 1); // Set pincer as target, i.e. set digital pin to 2
            //            }
            break;
          case 7:
            //            Serial.print(F("S2 pressed: "));
            //            Serial.print(pin_idx);
            //            if (2 == digital_pin) {
            //              res_locator(8, 4, 1); // execute sequence menu #8 item 4 alternate between servo extremes
            //            } else {
            //              res_locator(8, 7, 1); // Set left/right as target, i.e. set digital pin to 3
            //            }
            break;
          case 4:
            //            Serial.print(F("S3 pressed: "));
            //            Serial.print(pin_idx);
            //            if (2 == digital_pin) {
            //              res_locator(8, 4, 1); // execute sequence menu #8 item 4 alternate between servo extremes
            //            } else {
            //              res_locator(8, 8, 1); // Set up/down as target, i.e. set digital pin to 4
            //            }
            break;
          case 5:
            //            Serial.print(F("S4 pressed: "));
            //            Serial.print(pin_idx);
            //            if (2 == digital_pin) {
            //              res_locator(8, 4, 1); // execute sequence menu #8 item 4 alternate between servo extremes
            //            } else {
            //              res_locator(8, 9, 1); // Set out/in as target, i.e. set digital pin to 5
            //            }
            break;
        }
      }
    }
    pin_last[pin_idx] = pin_state[pin_idx];
  }
}

void processgtSerial1Data() {
}

void processgtSerial2Data() {
  if (rx_byte_S2 != 10) { // newline
    if (rx_byte_S2 > 31) {
      if (rx_byte_S2 < 127) {
        if (data_pos_S2 < 80) {
          cmdResponse_S2[data_pos_S2] = rx_byte_S2;
          data_pos_S2++;
        }
      }
    }
  }
  if (rx_byte_S2 == 13) {
    cmdResponse_S2[data_pos_S2] = '\0';
    data_pos_S2 = 0;
    if (show_responses > 0) {
      Serial.println();
      Serial.print(last_command);
      Serial.print(F(": "));
      Serial.print(cmdResponse_S2);
    }
    processResponse();
  }
}

void processResponse() {
  entered_value = 0;
  digit_count = 2;
  while (digit_count < 5) {
    entered_value = entered_value * 10 + last_command.charAt(digit_count) - 48;
    digit_count++;
    if (last_command.charAt(digit_count) < 48) {
      digit_count = 5;
    }
  }
  // the entered value in this case is the three-digit command
  switch (entered_value) {
    case 965: // pulse loop size
      cmdResponse_S2[0] = cmdResponse_S2 [4];
      cmdResponse_S2[1] = cmdResponse_S2 [5];
      cmdResponse_S2[2] = cmdResponse_S2 [6];
      cmdResponse_S2[3] = cmdResponse_S2 [7];
      cmdResponse_S2[4] = cmdResponse_S2 [8];
      numericResponse();
      parm_value_V = entered_value;
      break;
    case 975: // slew rate
      cmdResponse_S2[0] = cmdResponse_S2 [5];
      cmdResponse_S2[1] = cmdResponse_S2 [6];
      cmdResponse_S2[2] = cmdResponse_S2 [7];
      cmdResponse_S2[3] = cmdResponse_S2 [8];
      cmdResponse_S2[4] = cmdResponse_S2 [9];
      numericResponse();
      parm_value_W = entered_value;
      break;
    case 977: // max pulse width
      numericResponse();
      parm_value_Z = entered_value;
      switch (digital_pin) { // current digital pin #
        case 2:
          pin_2_max = entered_value;
          break;
        case 3:
          pin_3_max = entered_value;
          break;
        case 4:
          pin_4_max = entered_value;
          break;
        case 5:
          pin_5_max = entered_value;
          break;
      }
      break;
    case 978: // min pulse width
      numericResponse();
      parm_value_Y = entered_value;
      switch (digital_pin) { // current digital pin #
        case 2:
          pin_2_min = entered_value;
          break;
        case 3:
          pin_3_min = entered_value;
          break;
        case 4:
          pin_4_min = entered_value;
          break;
        case 5:
          pin_5_min = entered_value;
          break;
      }
      break;
    case 979: // target pulse width
      numericResponse();
      parm_value_X = entered_value;
      break;
    case 980: //
      numericResponse();
      pulseWidth = entered_value;
      break;
    case 984: //
      numericResponse();
      sourcePin = entered_value;
      break;
    case 985: //
      savResponse[11] = cmdResponse_S2 [0];
      savResponse[12] = cmdResponse_S2 [1];
      savResponse[13] = cmdResponse_S2 [2];
      savResponse[14] = cmdResponse_S2 [3];
      savResponse[15] = cmdResponse_S2 [4];
      savResponse[16] = cmdResponse_S2 [5];
      savResponse[17] = cmdResponse_S2 [6];
      savResponse[18] = cmdResponse_S2 [7];
      break;
    case 986: //
      savResponse[8] = cmdResponse_S2 [0];
      savResponse[9] = cmdResponse_S2 [1];
      savResponse[10] = cmdResponse_S2 [2];
      break;
    case 987: //
      savResponse[2] = cmdResponse_S2 [0];
      savResponse[3] = cmdResponse_S2 [1];
      savResponse[4] = cmdResponse_S2 [2];
      break;
    case 988: //
      savResponse[5] = cmdResponse_S2 [0];
      savResponse[6] = cmdResponse_S2 [1];
      savResponse[7] = cmdResponse_S2 [2];
      break;
    case 990: // current digital pin
      cmdResponse_S2[0] = cmdResponse_S2 [1];
      cmdResponse_S2[1] = cmdResponse_S2 [2];
      cmdResponse_S2[2] = cmdResponse_S2 [3];
      numericResponse();
      if (entered_value > 1) {
        if (entered_value < 14) {
          digital_pin = entered_value;
        }
      }
      break;
    default:
      if (entered_value > 99) {
        if (entered_value < 990) {
          Serial.print(F("command = S1"));
          Serial.println(entered_value);
          Serial.print(F(", response "));
          Serial.println(cmdResponse_S2);
          Serial.println(F(" was not saved."));
        }
      }
      break;
  }
}

void numericResponse() {
  entered_value = 0;
  digit_count = 0;
  while (digit_count < 5) {
    entered_value = entered_value * 10 + cmdResponse_S2[digit_count] - 48;
    digit_count++;
    if (cmdResponse_S2[digit_count] < 48) {
      digit_count = 5;
    }
  }
}

// each character is processed as it is entered
// a character may be processed differently according to the current value of input_mode.
void processCommand() {
  int command_number = recd_char;
  if (command_number > 96) {
    if (command_number < 123) {
      command_number -= 32; // treat upper and lower case the same
    }
  }
  switch (command_number) {
    case 3: // control-C
      terminateCurrentSeq();
      break;
    case 10: // newline (ignored)
      break;
    case 13: // carriage return
      captureValue();
      break;
    case 43: // +
      processIncrementKey();
      break;
    case 45: // -
      processDecrementKey();
      break;
    case 46: // . (period)
      captureValue();
      break;
    case 47: // / (forward slash)
      toggleViewSending();
      break;
    case 48:
    case 49:
    case 50:
    case 51:
    case 52:
    case 53:
    case 54:
    case 55:
    case 56:
    case 57:
      processDigitKey();
      break;
    case 63: // ? main help page
      setInputMode(1, 1);
      break;
    case 65: // A
      setInputMode(2, 0);
      break;
    case 67: // C
      setInputMode(201, 0);
      break;
    case 68: // D
      setInputMode(3, 0);
      break;
    case 69: // E - menu option key
      res_locator(menu_number, 10, 1); // res_type 1 to execute the command
      break;
    case 70: // F - menu option key
      res_locator(menu_number, 11, 1); // res_type 1 to execute the command
      break;
    case 71: // G - menu option key
      res_locator(menu_number, 12, 1); // res_type 1 to execute the command
      break;
    case 72: // H - menu option key
      res_locator(menu_number, 13, 1); // res_type 1 to execute the command
      break;
    case 76: // L - display the servo limits help page
      setInputMode(13, 1);
      break;
    case 77: // M - re-display the current menu
      input_mode = 0;
      display_menu = 1;
      break;
    case 78: // N - display the number entry help page
      setInputMode(12, 1);
      break;
    case 79: // O - display the operations help page
      setInputMode(11, 1);
      break;
    case 80: // P - display the parm values help page
      setInputMode(10, 1);
      break;
    case 81: // Q - quit sequencing
      terminateCurrentSeq();
      break;
    case 83: // S
      setInputMode(202, 0);
      break;
    case 86: // V
      setInputMode(4, 0);
      break;
    case 87: // W
      setInputMode(6, 0);
      break;
    case 88: // X
      setInputMode(7, 0);
      break;
    case 89: // Y
      setInputMode(8, 0);
      break;
    case 90: // Z
      setInputMode(9, 0);
      break;
    default:
      Serial.println();
      Serial.println();
      Serial.print(F("No command has been coded for this key press: "));
      Serial.print(recd_char);
      Serial.print(F(" ("));
      Serial.print(command_number);
      Serial.println(F(") "));
      break;
  }
}

void setInputMode(int new_mode, int help_mode) {
  if (help_mode > 0) {
    input_mode = 0;
  }
  if (input_mode != 0) {
    input_mode = 0; // revert to menu mode
  } else {
    if (new_mode > 201) {
      input_mode = 5; // input a sequence menu number
      if (menu_number < 8) {
        menu_number = seq_menu;
        input_mode = 0; // revert to menu mode
      }
    } else {
      if (new_mode > 200) {
        input_mode = 5; // input a command menu number
        if (menu_number > 7) {
          menu_number = cmd_menu;
          input_mode = 0; // revert to menu mode
        }
      } else {
        entered_value = 0;
        digit_count = 0;
        input_mode = new_mode; // value input mode
      }
    }
  }
  display_menu = 1;
}

void processIncrementKey() {
  switch (input_mode) {
    case 0: // command menu mode
      if (menu_number < 9) {
        menu_number++;
        if (menu_number < 8) {
          cmd_menu = menu_number;
        } else {
          seq_menu = menu_number;
        }
        display_menu = 1;
      }
      break;
    case 2: // analog pin mode
      if (analog_pin < 7) {
        analog_pin++;
        setCurrentAnalogPin();
        input_mode = 0;
        display_menu = 1;
      }
      break;
    case 3: //  digital pin mode
      if (digital_pin < 13) {
        digital_pin++;
        set_pin_digital();
        input_mode = 0;
        display_menu = 1;
      }
      break;
    case 4: // value mode
      if (parm_value_V + 1 > -1) {
        parm_value_V++;
        input_mode = 0;
        display_menu = 1;
      }
      break;
    case 6: // value mode
      if (parm_value_W + 1 > -1) {
        parm_value_W++;
        input_mode = 0;
        display_menu = 1;
      }
      break;
    case 7: // value mode
      if (parm_value_X + 1 > -1) {
        parm_value_X++;
        input_mode = 0;
        display_menu = 1;
      }
      break;
    case 8: // value mode
      if (parm_value_Y + 1 > -1) {
        parm_value_Y++;
        input_mode = 0;
        display_menu = 1;
      }
      break;
    case 9: // value mode
      if (parm_value_Z + 1 > -1) {
        parm_value_Z++;
        input_mode = 0;
        display_menu = 1;
      }
      break;
  }
}

void processDecrementKey() {
  switch (input_mode) {
    case 0: // menu mode (commands)
      if (menu_number > 0) {
        menu_number--;
        if (menu_number > 7) {
          seq_menu = menu_number;
        } else {
          cmd_menu = menu_number;
        }
        display_menu = 1;
      }
      break;
    case 2: // analog pin mode
      if (analog_pin > 0) {
        analog_pin--;
        setCurrentAnalogPin();
        input_mode = 0;
        display_menu = 1;
      }
      break;
    case 3: //  digital pin mode
      if (digital_pin > 2) {
        digital_pin--;
        set_pin_digital();
        input_mode = 0;
        display_menu = 1;
      }
      break;
    case 4: // value mode
      if (parm_value_V - 1 > -1) {
        parm_value_V--;
        input_mode = 0;
        display_menu = 1;
      }
      break;
    case 6: // value mode
      if (parm_value_W - 1 > -1) {
        parm_value_W--;
        input_mode = 0;
        display_menu = 1;
      }
      break;
    case 7: // value mode
      if (parm_value_X - 1 > -1) {
        parm_value_X--;
        input_mode = 0;
        display_menu = 1;
      }
      break;
    case 8: // value mode
      if (parm_value_Y - 1 > -1) {
        parm_value_Y--;
        input_mode = 0;
        display_menu = 1;
      }
      break;
    case 9: // value mode
      if (parm_value_Z - 1 > -1) {
        parm_value_Z--;
        input_mode = 0;
        display_menu = 1;
      }
      break;
  }
}

void processDigitKey() {
  digit_key = recd_char - 48;
  switch (input_mode) {
    case 2:
    case 3:
    case 4:
    case 6:
    case 7:
    case 8:
    case 9:
      // a value is being entered
      entered_value = entered_value * 10 + digit_key;
      digit_count++;
      break;
    case 5:
      // a menu is being selected
      if (digit_key > 0) {
        if (menu_number > 7) { // current menu type is "sequence"
          if (digit_key < 3) {
            menu_number = digit_key + 7; // sequence menu 1 is actual menu 8
            seq_menu = menu_number;
            display_menu = 1;
          }
        } else { // current menu type is "command"
          if (digit_key < 9) {
            menu_number = digit_key - 1; // sequence menu 1 is actual menu 0
            cmd_menu = menu_number;
            display_menu = 1;
          }
        }
      }
      input_mode = 0;
      break;
    default:
      // a menu option is being selected
      digit_key--;
      if (digit_key < 0) {
        digit_key += 10;
      }
      res_locator(menu_number, digit_key, 1); // res_type 1 to execute the command
      break;
  }
}

void captureValue() {
  if (digit_count > 0) {
    switch (input_mode) {
      case 2: // analog pin
        if (entered_value > -1) {
          if (entered_value < 8) {
            analog_pin = entered_value;
            setCurrentAnalogPin();
          }
        }
        input_mode = 0;
        display_menu = 1;
        break;
      case 3: // digital pin
        if (entered_value > 1) {
          if (entered_value < 14) {
            digital_pin = entered_value;
            set_pin_digital();
          }
        }
        input_mode = 0;
        display_menu = 1;
        break;
      case 4: // parm value
        setTempValue();
        parm_value_V = temp_value;
        break;
      case 6: // parm value
        setTempValue();
        parm_value_W = temp_value;
        break;
      case 7: // parm value
        setTempValue();
        parm_value_X = temp_value;
        break;
      case 8: // parm value
        setTempValue();
        parm_value_Y = temp_value;
        break;
      case 9: // parm value
        setTempValue();
        parm_value_Z = temp_value;
        break;
    }
  }
}

void setTempValue() {
  temp_value = inRange(entered_value, 0, 32766);
  input_mode = 0;
  display_menu = 1;
}

void setPinIO(int idx, int is_output) {
  pin_io[idx] = 0;
  pinMode(idx + 2, INPUT);
  if (is_output > 1) {
    pin_io[idx] = 1;
    pinMode(idx + 2, OUTPUT);
    digitalWrite(idx + 2, LOW);
  }
}

void consoleRefresh() {
  Serial.println();
  Serial.println();
  switch (input_mode) {
    case 0:
      consolePage_1();
      break;
    case 1:
      consolePage_2();
      break;
    case 2:
      analogPinMode();
      break;
    case 3:
      digitalPinMode();
      break;
    case 4:
      valueEntryMode(input_mode);
      break;
    case 5:
      menuJumpMode();
      break;
    case 6:
      valueEntryMode(input_mode);
      break;
    case 7:
      valueEntryMode(input_mode);
      break;
    case 8:
      valueEntryMode(input_mode);
      break;
    case 9:
      valueEntryMode(input_mode);
      break;
    case 10:
      consolePage_3();
      break;
    case 11:
      consolePage_4();
      break;
    case 12:
      consolePage_5();
      break;
    case 13:
      consolePage_6();
      break;
  }
}

void consolePage_1() {
  if (menu_number > 7) {
    Serial.print(F("Sequence Menu "));
    Serial.print(menu_number - 7);
  } else {
    Serial.print(F("Command Menu "));
    Serial.print(menu_number + 1);
  }
  Serial.print(F(" ("));
  Serial.print(menu_number);
  Serial.print(F(":"));
  Serial.print(commands_sent);
  Serial.println(F(")"));
  Serial.println();
  for (int line_number = 0; line_number < 14; line_number++) {
    if (line_number < 9) {
      Serial.print(line_number + 1);
    } else {
      switch (line_number) {
        case 9:
          Serial.print(F("0"));
          break;
        case 10:
          Serial.print(F("E"));
          break;
        case 11:
          Serial.print(F("F"));
          break;
        case 12:
          Serial.print(F("G"));
          break;
        case 13:
          Serial.print(F("H"));
          break;
      }
    }
    Serial.print(F(" - "));
    res_locator(menu_number, line_number, 0); // res_type 0 to print the command description
    Serial.println();
  }
  Serial.print(F("Other Commands: A D V W X Y Z C S M ? P Q"));
  requestText(F("M01"));
}

void consolePage_2() {
  Serial.println(F("Help Overview"));
  Serial.println();
  Serial.println(F("The command module responds to single-character commands. No CR is ever required."));
  Serial.println(F("Numher entry may be terminated by CR or by entering the '.' character."));
  Serial.println(F("Command '?' displays this command summary. Alpha commands are not case sensitive."));
  Serial.println(F("Enter the following characters for additional help."));
  Serial.println();
  Serial.println(F("Press L for help controlling servo limits."));
  Serial.println(F("Press N for help with number entry commands: 'A', 'D', 'V', 'W', 'X', 'Y', and 'Z'."));
  Serial.println(F("Press O for help with menu command options 1-9, 0, 'E', 'F', 'G', and 'H'."));
  Serial.println(F("Press P for process and menu display commands: 'Q', Ctrl-C, 'M', 'C', 'S', '+' and '-'."));
  input_mode = 0;
}

void consolePage_3() {
  Serial.println(F("Process and Menu Display Command Help Page"));
  Serial.println();
  //  Serial.println(F("Process control commands 'Q' and 'Ctrl-C' terminate a sequence or looping process."));
  //  Serial.println(F("The sequence process sends commands at timed intervals according to a script."));
  //  Serial.println(F("Scripted commands are preprogrammed and set in code and may loop or not as required."));
  //  Serial.println();
  //  Serial.println(F("Commands that control the menu display process do not themselves send commands to other nodes."));
  //  Serial.println(F("Command 'M' displays the current menu."));
  //  Serial.println(F("The 'C' and 'S' menu types are merely for convenience."));
  //  Serial.println(F("'C' menu options may initiate sequences and 'S' menu items may initiate simple commands."));
  //  Serial.println(F("If the current menu is a 'C' menu the 'C' command followed by a digit jumps to the indicated 'C' menu."));
  //  Serial.println(F("If the current menu is a 'C' menu the 'S' command switches to the last viewed sequence menu."));
  Serial.println(F("If the current menu is an 'S' menu the 'S' command followed by a digit jumps to that 'S' menu."));
  Serial.println(F("If the current menu is an 'S' menu the 'C' command switches to the last viewed command menu."));
  Serial.println(F("If not entering a value the '+' command will advance to the next menu."));
  Serial.println(F("If not entering a value the '-' command will move back to the previous menu."));
  input_mode = 0;
}

void consolePage_4() {
  Serial.println(F("Options Help Page"));
  Serial.println();
  Serial.println(F("Entry of a numeric character can go one of two ways. A value may be set or a menu option executed."));
  Serial.println(F("When not in number entry mode a numeric character executes an option from the current menu."));
  Serial.println(F("The letters 'E', 'F', 'G' and 'H' also execute options from the current menu."));
  Serial.println(F("This gives a total of up to 14 options on each menu."));
  Serial.println(F("Here is what happens when a menu option is selected."));
  Serial.println();
  Serial.println(F("The menu option description is displayed."));
  Serial.println(F("A command sequence is loaded into memory if the option initiates a sequence. If not ..."));
  Serial.println(F("The command is prepared."));
  Serial.println(F("The command is displayed. (this may be toggled on/off with '/')"));
  Serial.println(F("The command is sent."));
  Serial.println(F("If a sequence is in memory these three steps are repeated as the time elapses for each sequential step."));
  Serial.println();
  Serial.println(F("Replies from the remote node are displayed line by line as they are received."));
  Serial.println(F("Commands that change values in the target module do not (typically) cause a reply to be returned."));
  Serial.println(F("If confirmation of the desired change is needed a second option needs to be selected."));
  Serial.println(F("When a sequence is terminated the command sequence in memory is deleted."));
  Serial.println(F("If additional sequences are needed the Arduino sketch can be modified as required."));
  input_mode = 0;
}

void consolePage_5() {
  //  Serial.println(F("Number Entry Help Page"));
  //  Serial.println();
  //  Serial.println(F("Numeric characters are not treated as commands when entering values but are temporarily saved."));
  //  Serial.println(F("Commands 'A', 'D', 'V', 'W', 'X', 'Y', and 'Z' set values."));
  //  Serial.println(F("This continues until the number entry process is terminated."));
  //  Serial.println(F("If terminated by the '+' character the current pin or value is incremented."));
  //  Serial.println(F("If terminated by the '-' character the current pin or value is decremented."));
  //  Serial.println(F("If terminated by the . character or CR the number is stored in the variable."));
  //  Serial.println(F("Other commands can terminate the process and the variable will stay as it was."));
  //  Serial.println();
  //  Serial.println(F("These are the current values of the variables."));
  //  Serial.print(F("A = "));
  //  Serial.print(analog_pin);
  //  Serial.println(F(" (the current Analog Pin)"));
  //  Serial.print(F("D = "));
  //  Serial.print(digital_pin);
  //  Serial.print(F(","));
  //  prt_digit_pin();
  Serial.println();
  Serial.print(F("V = "));
  Serial.println(parm_value_V);
  Serial.print(F("W = "));
  Serial.println(parm_value_W);
  Serial.print(F("X = "));
  Serial.println(parm_value_X);
  Serial.print(F("Y = "));
  Serial.println(parm_value_Y);
  Serial.print(F("Z = "));
  Serial.println(parm_value_Z);
  input_mode = 0;
}

void consolePage_6() {
  Serial.println(F("Servo Limit Help Page"));
  Serial.println();
  Serial.println(F("Servo limits are values in the servo module that can limit a servo's range of motion."));
  Serial.println(F("Special commands help control and manipulate these values."));
  Serial.println(F("In the demo code values for servos on digital pins 2, 3, 4 and 5 are manipulated."));
  Serial.println(F("These are the current values of the variables."));
  Serial.println();
  limitReport();
  input_mode = 0;
}

void limitReport() {
  Serial.print(F("Servo on D2: Min = "));
  Serial.print(pin_2_min);
  Serial.print(F(", Max = "));
  Serial.println(pin_2_max);
  Serial.print(F("Servo on D3: Min = "));
  Serial.print(pin_3_min);
  Serial.print(F(", Max = "));
  Serial.println(pin_3_max);
  Serial.print(F("Servo on D4: Min = "));
  Serial.print(pin_4_min);
  Serial.print(F(", Max = "));
  Serial.println(pin_4_max);
  Serial.print(F("Servo on D5: Min = "));
  Serial.print(pin_5_min);
  Serial.print(F(", Max = "));
  Serial.println(pin_5_max);
}

void analogPinMode() {
  if (digit_count < 1) {
    Serial.print(F("Current analog Pin: "));
    Serial.println(analog_pin);
    Serial.print(F("Enter an analog pin #"));
    followedBy();
  } else {
    prtEnteredValue();
  }
}

void digitalPinMode() {
  if (digit_count < 1) {
    Serial.print(F("Current digital pin: "));
    Serial.println(digital_pin);
    Serial.print(F("Enter a digital pin #"));
    followedBy();
  } else {
    prtEnteredValue();
  }
}

void menuJumpMode() {
  if (digit_count < 1) {
    Serial.println(F("Enter a menu # or letter."));
  }
}

void valueEntryMode(int which_value) {
  if (digit_count < 1) {
    Serial.print(F("Current value: "));
    switch (which_value) {
      case 4:
        Serial.println(parm_value_V);
        break;
      case 6:
        Serial.println(parm_value_W);
        break;
      case 7:
        Serial.println(parm_value_X);
        break;
      case 8:
        Serial.println(parm_value_Y);
        break;
      case 9:
        Serial.println(parm_value_Z);
        break;
    }
    Serial.print(F("Enter a value"));
    followedBy();
  } else {
    prtEnteredValue();
  }
}

void followedBy() {
  Serial.print(F(" followed by . (period) or return."));
}

void prtEnteredValue() {
  Serial.println();
  Serial.print(F("Entered value: "));
  Serial.println(entered_value);
}

// mnu_num is menu number (zero based)
// lne_num is line number (zero based)
// res_type is 0 for command description only, 1 for describe and execute
void res_locator(int mnu_num, int lne_num, int res_type) {
  if (res_type > 0) {
    Serial.println();
    Serial.println();
  }
  switch (mnu_num) {
    case 0:
      menu_num_0_cmd_menu_1(lne_num, res_type);
      break;
    case 1:
      menu_num_1_cmd_menu_2(lne_num, res_type);
      break;
    case 2:
      menu_num_2_cmd_menu_3(lne_num, res_type);
      break;
    case 3:
      menu_num_3_cmd_menu_4(lne_num, res_type);
      break;
    case 4:
      menu_num_4_cmd_menu_5(lne_num, res_type);
      break;
    case 5:
      menu_num_5_cmd_menu_6(lne_num, res_type);
      break;
    case 6:
      menu_num_6_cmd_menu_7(lne_num, res_type);
      break;
    case 7:
      menu_num_7_cmd_menu_8_history(lne_num, res_type);
      break;
    case 8:
      menu_num_8_seq_menu_1(lne_num, res_type);
      break;
    case 9:
      menu_num_9_seq_menu_2(lne_num, res_type);
      break;
  }
  if (res_type > 0) {
    if (sequence_steps_to_do > 0) {
      if (show_responses > 0) { // show responses
        if (menu_option < 1) {
          Serial.println();
          Serial.print(F("Control-C or Q to quit sequencer"));
        }
      }
    }
  }
}

void menu_num_0_cmd_menu_1(int line_no, int cmd_type) {
  switch (line_no) {
    case 0:
      Serial.print(F("S1025 - Set all the digital IO to Input"));
      if (cmd_type > 0) {
        sendCommand(F("S1025"));
      }
      break;
    case 1:
      Serial.print(F("S1026 - Set the servo module to DEMO mode"));
      if (cmd_type > 0) {
        sendCommand(F("S1026"));
      }
      break;
    case 2:
      Serial.print(F("S1999 - Comprehensive Status Report"));
      if (cmd_type > 0) {
        sendCommand(F("S1999"));
      }
      break;
    case 3:
      Serial.print(F("S1998 - All Three Digital Reports"));
      if (cmd_type > 0) {
        sendCommand(F("S1998"));
      }
      break;
    case 4:
      Serial.print(F("S1997 - Analog Report - Pin Settings"));
      if (cmd_type > 0) {
        sendCommand(F("S1997"));
      }
      break;
    case 5:
      Serial.print(F("S1996 - Digital Report 1 - Pin Function Settings"));
      if (cmd_type > 0) {
        sendCommand(F("S1996"));
      }
      break;
    case 6:
      Serial.print(F("S1995 - Digital Report 2 - State, On/Off Times, Var 1/Var 2"));
      if (cmd_type > 0) {
        sendCommand(F("S1995"));
      }
      break;
    case 7:
      Serial.print(F("S1994 - Digital Report 3 - PWM Settings"));
      if (cmd_type > 0) {
        sendCommand(F("S1994"));
      }
      break;
    case 8:
      temp_value = inRange(parm_value_V, 5, 4000);
      Serial.print(F("S1024,"));
      Serial.print(fiveDigitParm(temp_value));
      Serial.print(F(" - Set the pulse loop size to "));
      Serial.print(temp_value);
      Serial.print(F(" (V: 5-4000)"));
      if (cmd_type > 0) {
        tmpString = F("S1024,");
        tmpString += fiveDigitParm(temp_value);
        sendCommand(tmpString);
      }
      break;
    case 9:
      Serial.print(F("S1965 - Retrieve the pulse loop size (V)"));
      if (cmd_type > 0) {
        sendCommand(F("S1965"));
      }
      break;
    case 10:
      Serial.print(F("Synchronize Command Module with Remote Node"));
      if (cmd_type > 0) {
        seq_prep(1); // hide responses
        set_step(100, F("S1965")); // get pulse loop size
        set_step(100, F("S1990")); // get current digital pin
        set_step(100, F("M0305")); // get servo limits
      }
      break;
    case 11: // F
      Serial.print(F("S1001 - Set mode to LOCAL"));
      if (cmd_type > 0) {
        sendCommand(F("S1001"));
      }
      break;
    case 12: // G
      Serial.print(F("S1002 - Set mode to REMOTE"));
      if (cmd_type > 0) {
        sendCommand(F("S1002"));
      }
      break;
    case 13: // H
      Serial.print(F("S1991 - Retrieve 'LOCAL' or 'REMOTE'"));
      if (cmd_type > 0) {
        sendCommand(F("S1991"));
      }
      break;
    default:
      no_command_assigned(0, line_no, cmd_type);
      break;
  }
}

void menu_num_1_cmd_menu_2(int line_no, int cmd_type) {
  switch (line_no) {
    case 0:
      Serial.print(F("S1993 - Get settings for"));
      prt_digit_pin();
      if (cmd_type > 0) {
        sendCommand(F("S1993"));
      }
      break;
    case 1:
      Serial.print(F("S1990 - Retrieve current digital pin #"));
      if (cmd_type > 0) {
        sendCommand(F("S1990"));
      }
      break;
    case 2:
      Serial.print(F("S1983 - Retrieve 0 or 1 for"));
      prt_digit_pin();
      if (cmd_type > 0) {
        sendCommand(F("S1983"));
      }
      break;
    case 3:
      temp_value = inRange(parm_value_V, 0, 4);
      Serial.print(F("S1007,"));
      Serial.print(twoDigitParm(temp_value));
      Serial.print(F(" - Set"));
      prt_digit_pin();
      Serial.print(F(" function to "));
      printFunction(temp_value);
      Serial.print(F(" (V: 0-4)"));
      if (cmd_type > 0) {
        tmpString = F("S1007,");
        tmpString += twoDigitParm(temp_value);
        sendCommand(tmpString);
      }
      break;
    case 4:
      Serial.print(F("S1987 - Retrieve 'RAW', 'DEB', 'LED' or 'PWM'"));
      if (cmd_type > 0) {
        sendCommand(F("S1987"));
      }
      break;
    case 5:
      Serial.print(F("S1005 - Lock"));
      prt_digit_pin();
      if (cmd_type > 0) {
        sendCommand(F("S1005"));
      }
      break;
    case 6:
      Serial.print(F("S1006 - Unlock"));
      prt_digit_pin();
      if (cmd_type > 0) {
        sendCommand(F("S1006"));
      }
      break;
    case 7:
      Serial.print(F("S1988 - Retrieve 'SWI' or 'LOC'"));
      if (cmd_type > 0) {
        sendCommand(F("S1988"));
      }
      break;
    case 8:
      Serial.print(F("S1986 - Retrieve 'IN' or 'OUT'"));
      if (cmd_type > 0) {
        sendCommand(F("S1986"));
      }
      break;
    case 9:
      temp_value = inRange(parm_value_W, 0, 5);
      Serial.print(F("S1008,"));
      Serial.print(twoDigitParm(temp_value));
      Serial.print(F(" - Set"));
      prt_digit_pin();
      Serial.print(F(" logic to "));
      printLogic(temp_value);
      Serial.print(F(" (W: 0-5)"));
      if (cmd_type > 0) {
        tmpString = F("S1008,");
        tmpString += twoDigitParm(temp_value);
        sendCommand(tmpString);
      }
      break;
    case 10:
      Serial.print(F("S1985 - Retrieve 'REM', 'BTN', 'POT', 'PMN' or 'PTR'"));
      if (cmd_type > 0) {
        sendCommand(F("S1985"));
      }
      break;
    case 11:
      temp_value = inRange(parm_value_X, 0, 13);
      Serial.print(F("S1009,"));
      Serial.print(twoDigitParm(temp_value));
      Serial.print(F(" - Set the control input pin for"));
      prt_digit_pin();
      Serial.print(F("to "));
      Serial.print(temp_value);
      Serial.print(F(" (X: 0-13)"));
      if (cmd_type > 0) {
        tmpString = F("S1009,");
        tmpString += twoDigitParm(temp_value);
        sendCommand(tmpString);
      }
      break;
    case 12:
      Serial.print(F("S1984 - Retrieve PARM source pin for"));
      prt_digit_pin();
      if (cmd_type > 0) {
        sendCommand(F("S1984"));
      }
      break;
    default:
      no_command_assigned(1, line_no, cmd_type);
      break;
  }
}

void menu_num_2_cmd_menu_3(int line_no, int cmd_type) {
  switch (line_no) {
    case 0:
      Serial.print(F("S1993 - Get settings for"));
      prt_digit_pin();
      if (cmd_type > 0) {
        sendCommand(F("S1993"));
      }
      break;
    case 1:
      Serial.print(F("S1015 - Set the direction to REVERSE"));
      if (cmd_type > 0) {
        sendCommand(F("S1015"));
      }
      break;
    case 2:
      Serial.print(F("S1016 - Set the direction to NORMAL"));
      if (cmd_type > 0) {
        sendCommand(F("S1016"));
      }
      break;
    case 3:
      Serial.print(F("S1974 - Retrieve 'NORMAL' or 'REVERSED'"));
      if (cmd_type > 0) {
        sendCommand(F("S1974"));
      }
      break;
    case 4:
      temp_value = inRange(parm_value_W, 0, 300);
      Serial.print(F("S1014,"));
      Serial.print(fiveDigitParm(temp_value));
      Serial.print(F(" - Set the servo slew rate (W)"));
      if (cmd_type > 0) {
        tmpString = F("S1014,");
        tmpString += fiveDigitParm(temp_value);
        sendCommand(tmpString);
      }
      break;
    case 5:
      Serial.print(F("S1975 - Retrieve the current slew rate"));
      if (cmd_type > 0) {
        sendCommand(F("S1975"));
      }
      break;
    case 6:
      Serial.print(F("S1980 - Retrieve the current servo pulse width"));
      if (cmd_type > 0) {
        sendCommand(F("S1980"));
      }
      break;
    case 7:
      temp_value = inRange(parm_value_X, 100, 3000);
      Serial.print(F("S1011,"));
      Serial.print(fiveDigitParm(temp_value));
      Serial.print(F(" - Set the servo target (X)"));
      if (cmd_type > 0) {
        tmpString = F("S1011,");
        tmpString += fiveDigitParm(temp_value);
        sendCommand(tmpString);
      }
      break;
    case 8:
      Serial.print(F("S1979 - Retrieve the current target pulse width"));
      if (cmd_type > 0) {
        sendCommand(F("S1979"));
      }
      break;
    case 13:
      Serial.print(F("S1976 - Retrieve the current pulse width range"));
      if (cmd_type > 0) {
        sendCommand(F("S1976"));
      }
      break;
    default:
      no_command_assigned(4, line_no, cmd_type);
      break;
  }
}

void menu_num_3_cmd_menu_4(int line_no, int cmd_type) {
  switch (line_no) {
    case 0:
      Serial.print(F("S1993 - Get settings for"));
      prt_digit_pin();
      if (cmd_type > 0) {
        sendCommand(F("S1993"));
      }
      break;
    case 1:
      Serial.print(F("S1003,02 - Set pincer as target"));
      if (cmd_type > 0) {
        digital_pin = 2;
        set_pin_digital();
      }
      break;
    case 2:
      Serial.print(F("S1003,03 - Set left/right as target"));
      if (cmd_type > 0) {
        digital_pin = 3;
        set_pin_digital();
      }
      break;
    case 3:
      Serial.print(F("S1003,04 - Set up/down as target"));
      if (cmd_type > 0) {
        digital_pin = 4;
        set_pin_digital();
      }
      break;
    case 4:
      Serial.print(F("S1003,05 - Set out/in as target"));
      if (cmd_type > 0) {
        digital_pin = 5;
        set_pin_digital();
      }
      break;
    case 5:
      Serial.print(F("Get Min and Max for demo servos"));
      if (cmd_type > 0) {
        seq_prep(1); // show responses
        set_step(100, F(">"));     // save current digital pin
        set_step(100, F("M0304")); // set out/in as target
        set_step(100, F("S1978")); // get min pulse width
        set_step(100, F("S1977")); // get max pulse width
        set_step(100, F("M0303")); // set up/down as target
        set_step(100, F("S1978")); // get min pulse width
        set_step(100, F("S1977")); // get max pulse width
        set_step(100, F("M0302")); // set left/right as target
        set_step(100, F("S1978")); // get min pulse width
        set_step(100, F("S1977")); // get max pulse width
        set_step(100, F("M0301")); // set pincer as target
        set_step(100, F("S1978")); // get min pulse width
        set_step(100, F("S1977")); // get max pulse width
        set_step(100, F("<"));     // restore current digital pin
      }
      break;
    case 6:
      temp_value = inRange(parm_value_Y, 100, 3000);
      Serial.print(F("S1012,"));
      Serial.print(fiveDigitParm(temp_value));
      Serial.print(F(" - Set the min servo pulse width (Y)"));
      if (cmd_type > 0) {
        tmpString = F("S1012,");
        tmpString += fiveDigitParm(temp_value);
        sendCommand(tmpString);
      }
      break;
    case 7:
      Serial.print(F("S1978 - Retrieve the current min pulse width (Y)"));
      if (cmd_type > 0) {
        sendCommand(F("S1978"));
      }
      break;
    case 8:
      temp_value = inRange(parm_value_Z, 300, 3000);
      Serial.print(F("S1013,"));
      Serial.print(fiveDigitParm(temp_value));
      Serial.print(F(" - Set the max servo pulse width (Z)"));
      if (cmd_type > 0) {
        tmpString = F("S1013,");
        tmpString += fiveDigitParm(temp_value);
        sendCommand(tmpString);
      }
      break;
    case 9:
      Serial.print(F("S1977 - Retrieve the current max pulse width (Z)"));
      if (cmd_type > 0) {
        sendCommand(F("S1977"));
      }
      break;
    case 10:
      Serial.print(F("Move Servo cconnected to D"));
      Serial.print(digital_pin);
      if (cmd_type > 0) {
        seq_prep(1); // show responses
        set_step(100, F("S1008,00"));  // use REMOTE control
        set_step(100, F("S1011,01200")); // set servo target
        if (2 == digital_pin) {
          set_step(3999, F("S1011,00800")); // set servo target
        } else {
          set_step(3999, F("S1011,02000")); // set servo target
        }
        set_step(3999, F("S1011,01200")); // set servo target
        set_step(3899, F("R0205")); // return 2 steps
      }
      break;
    case 11:
      Serial.print(F("Motion Combo #1"));
      if (cmd_type > 0) {
        seq_prep(1); // show responses
        set_step(100, F("M0304")); // set out/in as target
        set_step(100, F("S1008,00"));  // use REMOTE control
        set_step(100, F("S1011,01200")); // set servo target
        set_step(100, F("M0303")); // set up/down as target
        set_step(100, F("S1008,00"));  // use REMOTE control
        set_step(100, F("S1011,01200")); // set servo target
        set_step(3999, F("M0304")); // set out/in as target
        set_step(100, F("S1011,02000")); // set servo target
        set_step(100, F("M0303")); // set up/down as target
        set_step(100, F("S1011,02000")); // set servo target
        set_step(3899, F("R1005")); // return 10 steps
      }
      break;
    case 12:
      Serial.print(F("Motion Combo #2"));
      if (cmd_type > 0) {
        seq_prep(1); // show responses
        set_step(100, F("M0302")); // set left/right as target
        set_step(100, F("S1008,00"));  // use REMOTE control
        set_step(100, F("S1011,01200")); // set servo target
        set_step(100, F("M0303")); // set up/down as target
        set_step(100, F("S1008,00"));  // use REMOTE control
        set_step(100, F("S1011,01200")); // set servo target
        set_step(3999, F("M0302")); // set left/right as target
        set_step(100, F("S1011,02000")); // set servo target
        set_step(100, F("M0303")); // set up/down as target
        set_step(100, F("S1011,02000")); // set servo target
        set_step(3899, F("R1005")); // return 10 steps
      }
      break;
    default:
      no_command_assigned(2, line_no, cmd_type);
      break;
  }
}

void menu_num_4_cmd_menu_5(int line_no, int cmd_type) {
  switch (line_no) {
    case 6:
      Serial.print(F("Send M02"));
      if (cmd_type > 0) {
        requestText(F("M02"));
      }
      break;
    case 7:
      Serial.print(F("Send M03"));
      if (cmd_type > 0) {
        requestText(F("M03"));
      }
      break;
    case 8:
      Serial.print(F("Send M04"));
      if (cmd_type > 0) {
        requestText(F("M04"));
      }
      break;
    case 9:
      Serial.print(F("Send M05"));
      if (cmd_type > 0) {
        requestText(F("M05"));
      }
      break;
    default:
      no_command_assigned(3, line_no, cmd_type);
      break;
  }
}

void menu_num_5_cmd_menu_6(int line_no, int cmd_type) {
  switch (line_no) {
    case 0:
      Serial.print(F("S1993 - Get settings for"));
      prt_digit_pin();
      if (cmd_type > 0) {
        sendCommand(F("S1993"));
      }
      break;
    case 1:
      Serial.print(F("S1017 - Set the trigger on"));
      prt_digit_pin();
      if (cmd_type > 0) {
        sendCommand(F("S1017"));
      }
      break;
    case 2:
      Serial.print(F("S1973 - Retrieve the current toggle state"));
      if (cmd_type > 0) {
        sendCommand(F("S1973"));
      }
      break;
    case 3:
      Serial.print(F("S1027 - Turn tracking on"));
      if (cmd_type > 0) {
        sendCommand(F("S1027"));
      }
      break;
    case 4:
      Serial.print(F("S1028 - Turn tracking off"));
      if (cmd_type > 0) {
        sendCommand(F("S1028"));
      }
      break;
    case 5:
      Serial.print(F("S1964 - Return tracker state"));
      if (cmd_type > 0) {
        sendCommand(F("S1964"));
      }
      break;
    case 6:
      temp_value = inRange(parm_value_Y, 100, 3000);
      Serial.print(F("S1010,"));
      Serial.print(fiveDigitParm(temp_value));
      Serial.print(F(" - Set the LED brightness for"));
      prt_digit_pin();
      Serial.print(F(" (X)"));
      if (cmd_type > 0) {
        tmpString = F("S1010,");
        tmpString += fiveDigitParm(temp_value);
        sendCommand(tmpString);
      }
      break;
    case 7:
      Serial.print(F("S1982 - Retrieve on-time for"));
      prt_digit_pin();
      if (cmd_type > 0) {
        sendCommand(F("S1982"));
      }
      break;
    case 8:
      Serial.print(F("S1981 - Retrieve off-time for"));
      prt_digit_pin();
      if (cmd_type > 0) {
        sendCommand(F("S1981"));
      }
      break;
    case 9:
      temp_value = inRange(parm_value_Y, 100, 3000);
      Serial.print(F("S1018,"));
      Serial.print(fiveDigitParm(temp_value));
      Serial.print(F(" - Set digital var #1 (Y)"));
      if (cmd_type > 0) {
        tmpString = F("S1018,");
        tmpString += fiveDigitParm(temp_value);
        sendCommand(tmpString);
      }
      break;
    case 10:
      Serial.print(F("S1972 - Retrieve digital var #1"));
      if (cmd_type > 0) {
        sendCommand(F("S1972"));
      }
      break;
    case 11:
      temp_value = inRange(parm_value_Y, 100, 3000);
      Serial.print(F("S1019,"));
      Serial.print(fiveDigitParm(temp_value));
      Serial.print(F(" - Set digital var #2 (Z)"));
      if (cmd_type > 0) {
        tmpString = F("S1019,");
        tmpString += fiveDigitParm(temp_value);
        sendCommand(tmpString);
      }
      break;
    case 12:
      Serial.print(F("S1971 - Retrieve digital var #2"));
      if (cmd_type > 0) {
        sendCommand(F("S1971"));
      }
      break;
    default:
      no_command_assigned(5, line_no, cmd_type);
      break;
  }
}

void menu_num_6_cmd_menu_7(int line_no, int cmd_type) {
  switch (line_no) {
    case 0:
      Serial.print(F("S1992 - Current Analog Pin Settings"));
      if (cmd_type > 0) {
        sendCommand(F("S1992"));
      }
      break;
    case 1:
      Serial.print(F("S1989 - Retrieve the current analog pin #"));
      if (cmd_type > 0) {
        sendCommand(F("S1989"));
      }
      break;
    case 2:
      Serial.print(F("S1970 - Retrieve the analog input value"));
      if (cmd_type > 0) {
        sendCommand(F("S1970"));
      }
      break;
    case 3:
      temp_value = inRange(parm_value_W, 300, 3000);
      Serial.print(F("S1020,"));
      Serial.print(fiveDigitParm(temp_value));
      Serial.print(F(" - Set the analog lower bound (W)"));
      if (cmd_type > 0) {
        tmpString = F("S1020,");
        tmpString += fiveDigitParm(temp_value);
        sendCommand(tmpString);
      }
      break;
    case 4:
      Serial.print(F("S1969 - Retrieve the lower bound"));
      if (cmd_type > 0) {
        sendCommand(F("S1969"));
      }
      break;
    case 5:
      temp_value = inRange(parm_value_X, 300, 3000);
      Serial.print(F("S1021,"));
      Serial.print(fiveDigitParm(temp_value));
      Serial.print(F(" - Set the analog upper bound (X)"));
      if (cmd_type > 0) {
        tmpString = F("S1021,");
        tmpString += fiveDigitParm(temp_value);
        sendCommand(tmpString);
      }
      break;
    case 6:
      Serial.print(F("S1968 - Retrieve the upper bound"));
      if (cmd_type > 0) {
        sendCommand(F("S1968"));
      }
      break;
    case 7:
      Serial.print(F("S1967 - Retrieve the range"));
      if (cmd_type > 0) {
        sendCommand(F("S1967"));
      }
      break;
    case 8:
      Serial.print(F("S1966 - Retrieve the scaled input value"));
      if (cmd_type > 0) {
        sendCommand(F("S1966"));
      }
      break;
    case 9:
      temp_value = inRange(parm_value_Y, 300, 3000);
      Serial.print(F("S1022,"));
      Serial.print(fiveDigitParm(temp_value));
      Serial.print(F(" - Set analog var #1 (Y)"));
      if (cmd_type > 0) {
        tmpString = F("S1022,");
        tmpString += fiveDigitParm(temp_value);
        sendCommand(tmpString);
      }
      break;
    case 10:
      Serial.print(F("S1963 - Retrieve analog var #1"));
      if (cmd_type > 0) {
        sendCommand(F("S1963"));
      }
      break;
    case 11:
      temp_value = inRange(parm_value_Z, 300, 3000);
      Serial.print(F("S1023,"));
      Serial.print(fiveDigitParm(temp_value));
      Serial.print(F(" - Set analog var #2 (Z)"));
      if (cmd_type > 0) {
        tmpString = F("S1023,");
        tmpString += fiveDigitParm(temp_value);
        sendCommand(tmpString);
      }
      break;
    case 12:
      Serial.print(F("S1962 - Retrieve analog var #2"));
      if (cmd_type > 0) {
        sendCommand(F("S1962"));
      }
      break;
    case 13:
      Serial.print(F("S1029 - Toggle analog toggle state"));
      if (cmd_type > 0) {
        sendCommand(F("S1029"));
      }
      break;
    default:
      no_command_assigned(6, line_no, cmd_type);
      break;
  }
}

void no_command_assigned(int menu, int key, int cmd_type) {
  if (cmd_type > 0) {
    Serial.print(F("Command Menu "));
    Serial.print(menu + 1);
    optionNotDefined(key);
  }
}

void output_print_buffer(int line_no) {
  if (strlen(outputBuffer[line_no]) > 0) {
    Serial.print(F("buffer line "));
    Serial.print(line_no);
    Serial.print(F(", len "));
    Serial.print(strlen(outputBuffer[line_no]));
    Serial.print(F(", "));
    Serial.print(outputBuffer[line_no]);
  }
}

void menu_num_7_cmd_menu_8_history(int line_no, int cmd_type) {
  if (strlen(cmdHistory[line_no]) > 0) {
    Serial.print(F("Send ("));
    Serial.print(strlen(cmdHistory[line_no]));
    Serial.print(F(") "));
    Serial.print(cmdHistory[line_no]);
    if (cmd_type > 0) {
      sendCommand(cmdHistory[line_no]);
    }
  }
}

void menu_num_8_seq_menu_1(int line_no, int cmd_type) {
  switch (line_no) {
    case 1:
      Serial.print(F("Set Pulse Loop Size"));
      if (cmd_type > 0) {
        seq_prep(1); // show responses
        set_step(200, F("S1024,00005"));
        set_step(200, F("S1024,00010"));
        set_step(200, F("S1024,00050"));
        set_step(200, F("S1024,00100"));
        set_step(200, F("S1024,00150"));
        set_step(200, F("S1024,00025"));
        set_step(200, F("S1965"));
      }
      break;
    case 2:
      Serial.print(F("Set Servo slew rate for D"));
      Serial.print(digital_pin);
      if (cmd_type > 0) {
        seq_prep(1); // show responses
        set_step(200, F("S1975"));
        set_step(100, F("S1014,00000")); // set slew rate
        set_step(100, F("S1014,00001")); // set slew rate
        set_step(100, F("S1014,00002")); // set slew rate
        set_step(100, F("S1014,00004")); // set slew rate
        set_step(100, F("S1014,00008")); // set slew rate
        set_step(100, F("S1014,00012")); // set slew rate
        set_step(200, F("S1975"));
      }
      break;
    case 3:
      Serial.print(F("Get settings for"));
      prt_digit_pin();
      if (cmd_type > 0) {
        seq_prep(0); // hide responses
        set_step(100, F("S1990"));
        set_step(100, F("S1987"));
        set_step(100, F("S1988"));
        set_step(100, F("S1986"));
        set_step(100, F("S1985"));
        set_step(100, F("S1984"));
        set_step(100, F("S1980"));
        set_step(100, F("S1979"));
        set_step(100, F("S1975"));
        set_step(100, F("S1028"));
        set_step(100, F("P1"));
      }
      break;
    default:
      if (cmd_type > 0) {
        Serial.print(F("Sequence Menu 1"));
        optionNotDefined(line_no);
      }
      break;
  }
}

void menu_num_9_seq_menu_2(int line_no, int cmd_type) {
  switch (line_no) {
    case 0:
      Serial.print(F("Blink LED cconnected to D"));
      Serial.print(digital_pin);
      if (cmd_type > 0) {
        seq_prep(1); // show responses
        set_step(100, F("S1006"));     // unlock pin
        set_step(100, F("S1007,02"));  // make it an LED output
        set_step(100, F("S1008,00"));  // use REMOTE control
        set_step(100, F("S1010,00100")); // turn LED on
        set_step(999, F("S1010,00000")); // turn LED off
        set_step(999, F("S1010,00100")); // turn LED on
        set_step(10, F("R0300")); // return 3 steps
      }
      break;
    case 6:
      Serial.print(F("S1003,09 - Set LED 1 as target"));
      if (cmd_type > 0) {
        digital_pin = 9;
        set_pin_digital();
        display_menu = 1;
      }
      break;
    case 7:
      Serial.print(F("S1003,11 - Set LED 2 as target"));
      if (cmd_type > 0) {
        digital_pin = 11;
        set_pin_digital();
        display_menu = 1;
      }
      break;
    case 8:
      Serial.print(F("S1003,13 - Set LED 3 as target"));
      if (cmd_type > 0) {
        digital_pin = 13;
        set_pin_digital();
        display_menu = 1;
      }
      break;
    case 9:
      Serial.print(F("S1003,10 - Set LED 4 as target"));
      if (cmd_type > 0) {
        digital_pin = 10;
        set_pin_digital();
        display_menu = 1;
      }
      break;
    case 10:
      Serial.print(F("S1003,8 - Set LED 5 as target"));
      if (cmd_type > 0) {
        digital_pin = 8;
        set_pin_digital();
        display_menu = 1;
      }
      break;
    case 11:
      Serial.print(F("S1003,12 - Set LED 6 as target"));
      if (cmd_type > 0) {
        digital_pin = 12;
        set_pin_digital();
        display_menu = 1;
      }
      break;
    case 12:
      Serial.print(F("S1003,07 - Set LED 7 as target"));
      if (cmd_type > 0) {
        digital_pin = 7;
        set_pin_digital();
        display_menu = 1;
      }
      break;
    default:
      if (cmd_type > 0) {
        Serial.print(F("Sequence Menu 2"));
        optionNotDefined(line_no);
      }
      break;
  }
}

void set_pin_digital() {
  temp_value = inRange(digital_pin, 2, 13);
  tmpString = F("S1003,");
  tmpString += twoDigitParm(temp_value);
  sendCommand(tmpString);
}

void prt_digit_pin() {
  Serial.print(F(" digital pin D"));
  Serial.print(digital_pin);
}

void optionNotDefined(int option_number) {
  Serial.print(F(" Option "));
  option_number++;
  switch (option_number) {
    case 10:
      Serial.print(F("0"));
      break;
    case 11:
      Serial.print(F("E"));
      break;
    case 12:
      Serial.print(F("F"));
      break;
    case 13:
      Serial.print(F("G"));
      break;
    case 14:
      Serial.print(F("H"));
      break;
    default:
      Serial.print(option_number);
      break;
  }
  Serial.print(F(" was selected but not defined"));
}

void printFunction(int fnc_number) {
  switch (fnc_number) {
    case 0:
      Serial.print(F("RAW"));
      break;
    case 1:
      Serial.print(F("DEB"));
      break;
    case 2:
      Serial.print(F("LED"));
      break;
    case 3:
      Serial.print(F("PWM"));
      break;
    case 4:
      Serial.print(F("PWA"));
      break;
  }
}

void printLogic(int logic_number) {
  switch (logic_number) {
    case 0:
      Serial.print(F("REM"));
      break;
    case 1:
      Serial.print(F("BTN"));
      break;
    case 2:
      Serial.print(F("POT"));
      break;
    case 3:
      Serial.print(F("PMN"));
      break;
    case 4:
      Serial.print(F("PTR"));
      break;
    case 5:
      Serial.print(F("PTA"));
      break;
  }
}

void setCurrentAnalogPin() {
  temp_value = inRange(analog_pin, 0, 7);
  tmpString = F("S1004,");
  tmpString += twoDigitParm(temp_value);
  sendCommand(tmpString);
}

void terminateCurrentSeq() {
  if (sequence_steps_to_do > 0) {
    sequence_steps_to_do = 0;
    show_responses = 1; // show responses
    input_mode = 0;
    Serial.println();
    Serial.println(F("Sequence terminated"));
  }
}

void seq_prep(int show) {
  step_no = 0;
  show_responses = show;
}

// this function merely stores the steps in an array
// for more details look at executeCommand
void set_step(int cmd_delay, const String & cmd_str) {
  if (step_no < NUMBER_OF_ELEMENTS) {
    cmd_str.toCharArray(seqCommand[step_no], 12);
    seqDelay[step_no] = cmd_delay;
    step_no++;
    sequence_steps_to_do = step_no;
    seq_step = 0;
  } else {
    Serial.println(F("It seems there are too many steps defined for this sequence"));
  }
}

void report_one() {
  Serial.println(F("Report 0ne"));
  Serial.println();
  Serial.print(F("Pin D"));
  Serial.print(digital_pin);
  Serial.print(F(", Function: "));
  Serial.print(savResponse[2]);
  Serial.print(savResponse[3]);
  Serial.print(savResponse[4]);
  Serial.print(F(", SWI/LOC: "));
  Serial.print(savResponse[5]);
  Serial.print(savResponse[6]);
  Serial.print(savResponse[7]);
  Serial.print(F(", Direction: "));
  Serial.print(savResponse[8]);
  Serial.print(savResponse[9]);
  Serial.println(savResponse[10]);
  Serial.print(F("Control: "));
  Serial.print(savResponse[11]);
  Serial.print(savResponse[12]);
  Serial.print(savResponse[13]);
  Serial.print(savResponse[14]);
  Serial.print(savResponse[15]);
  Serial.print(savResponse[16]);
  Serial.print(savResponse[17]);
  Serial.print(savResponse[18]);
  Serial.print(F(", Source Pin: "));
  Serial.println(sourcePin);
  Serial.print(F("Current Pulse Width: "));
  Serial.print(pulseWidth);
  Serial.print(F(", Target Pulse Width: "));
  Serial.println(parm_value_X);
  Serial.print(F("Slew Rate "));
  Serial.println(parm_value_W);
}

void report_two() {
  Serial.println(F("Report Two"));
  Serial.println();
  limitReport();
  Serial.println();
}

void sendCommand(String command_string) {
  myStr = "";
  int isDuplicate = 0;
  for (h_idx = 0; h_idx < 14; h_idx++) {
    myStr = "";
    myStr.concat(cmdHistory[h_idx]);
    if (command_string.equals(myStr)) {
      isDuplicate = 1;
      h_idx = 14;
    }
  }
  if (isDuplicate < 1) {
    for (h_idx = 0; h_idx < 13; h_idx++) {
      myStr = "";
      myStr.concat(cmdHistory[h_idx + 1]);
      myStr.toCharArray(cmdHistory[h_idx], 12);
    }
    command_string.toCharArray(cmdHistory[13], 12);
  }
  if (show_responses > 0) { // show responses
    if (view_sending > 0) {
      Serial.println();
      Serial.print(F("Sending: "));
      Serial.print(command_string);
    }
  }
  last_command = command_string;
  for (int i = 0; i < command_string.length(); i++) {
    gtSerial2.print(command_string.charAt(i));
  }
  gtSerial2.println();
  commands_sent++;
}

void requestText(String command_string) {
  //  Serial.println();
  //  Serial.print(F("requesting "));
  //  Serial.println(command_string);
  new_message = 1;
  for (int i = 0; i < command_string.length(); i++) {
    gtSerial1.print(command_string.charAt(i));
  }
  gtSerial1.println();
  message_mode = 1;
}

void toggleViewSending() {
  Serial.println();
  Serial.print(F("Sending: "));
  if (view_sending > 0) {
    view_sending = 0;
    Serial.print(F("hid"));
  } else {
    view_sending = 1;
    Serial.print(F("revealed"));
  }
}

String fiveDigitParm(int value) {
  String retString = "";
  if (value < 0) {
    value = 0;
  }
  if (value < 10000) {
    retString += F("0");
  }
  if (value < 1000) {
    retString += F("0");
  }
  if (value < 100) {
    retString += F("0");
  }
  if (value < 10) {
    retString += F("0");
  }
  retString += value;
  return retString;
}

int inRange(int value, int min_val, int max_val) {
  if (value < min_val) {
    value = min_val;
  }
  if (value > max_val) {
    value = max_val;
  }
  return value;
}

String twoDigitParm(int value) {
  String retString = "";
  if (value < 10) {
    retString += F("0");
  }
  retString += value;
  return retString;
}
